""" A Serializable Dataclass Implementation """
from __future__ import annotations

import dataclasses
import functools
import os
from typing import Any, Dict, Optional, Type, TypeVar, get_type_hints
from any_serde.common import (
    JSON,
    InvalidDeserializationException,
)


ATTR_SERIALIZATION_RENAMES = "__serialization_renames__"
ATTR_UNKNOWN_DATA_KEYS_POLICY = "__unknown_data_keys_policy__"
ATTR_INJECTED_DATA = "__injected_data__"


T_Dataclass = TypeVar("T_Dataclass", bound=object)


def is_dataclass_type(typ: Any) -> bool:
    return isinstance(typ, type) and dataclasses.is_dataclass(typ)


@functools.lru_cache(None)
def _get_type_hints(typ: Type[object]) -> Dict[str, Type[Any]]:
    return get_type_hints(typ)


def _field_is_required(field: dataclasses.Field) -> bool:
    if not isinstance(field.default, dataclasses._MISSING_TYPE):
        return False

    if not isinstance(field.default_factory, dataclasses._MISSING_TYPE):
        return False

    return True


@functools.lru_cache(None)
def _get_serialization_renames(dataclass_type: Type[T_Dataclass]) -> Dict[str, str]:
    error_msg_start = f"Illegal {ATTR_SERIALIZATION_RENAMES} value for {dataclass_type}!"

    serialization_renames = getattr(dataclass_type, ATTR_SERIALIZATION_RENAMES, {})
    if not isinstance(serialization_renames, dict):
        raise AssertionError(f"{error_msg_start} Not a dictionary!")

    for attribute_key, data_key in serialization_renames.items():
        if not isinstance(attribute_key, str) or not isinstance(data_key, str):
            raise AssertionError(f"{error_msg_start} Found non-string values!")

    attribute_keys = set(serialization_renames)
    data_keys = set(serialization_renames.values())
    if len(data_keys) < len(attribute_keys):
        raise AssertionError(f"{error_msg_start} Duplicate data keys!")

    dataclass_field_names = {field.name for field in dataclasses.fields(dataclass_type)}  # type: ignore[arg-type]
    if attribute_keys - dataclass_field_names:
        raise AssertionError(f"{error_msg_start} Mapping attributes that don't exist!")

    unmapped_attribute_keys = dataclass_field_names - attribute_keys
    if unmapped_attribute_keys & data_keys:
        raise AssertionError(f"{error_msg_start} Data keys conflict with existing fields!")

    return serialization_renames


@functools.lru_cache(None)
def _get_unknown_data_keys_policy(dataclass_type: Type[T_Dataclass]) -> UnknownDataKeysPolicy:
    if not hasattr(dataclass_type, ATTR_UNKNOWN_DATA_KEYS_POLICY):
        return UnknownDataKeysPolicy(
            allow=False,
            roundtrip=False,
        )

    policy = getattr(dataclass_type, ATTR_UNKNOWN_DATA_KEYS_POLICY)

    assert isinstance(policy, UnknownDataKeysPolicy), "Bad data keys policy"

    return policy


def from_data(type_: Type[T_Dataclass], data: JSON) -> T_Dataclass:
    if not isinstance(data, dict):
        raise InvalidDeserializationException(f"Dataclasses serialize to dict. Got {type(data)} instead!")

    assert is_dataclass_type(type_), "Can only call dataclass_serde.from_data on dataclass types!"
    injected_data: Optional[Dict[str, JSON]] = None

    field_types = _get_type_hints(type_)  # type: ignore
    dataclass_fields = {x.name: x for x in dataclasses.fields(type_)}  # type: ignore

    fields_without_types = set(dataclass_fields) - set(field_types)
    assert not fields_without_types, f"Fields without types: {fields_without_types}"

    serialization_renames = _get_serialization_renames(type_)  # type: ignore[arg-type]
    deserialization_renames = {data_key: attribute_key for attribute_key, data_key in serialization_renames.items()}
    mapped_data = {deserialization_renames.get(data_key, data_key): data_value for data_key, data_value in data.items()}

    data_keys_without_fields = set(mapped_data) - set(dataclass_fields)
    if data_keys_without_fields:
        unknown_data_keys_policy = _get_unknown_data_keys_policy(type_)  # type: ignore[arg-type]
        if not unknown_data_keys_policy.allow:
            # TODO: not intuitive that these printed are after mapping. Either print before map or explain.
            raise InvalidDeserializationException(
                f"Parsing {type_}, got data keys without fields: {data_keys_without_fields}"
            )

        if unknown_data_keys_policy.roundtrip:
            injected_data = {data_key: mapped_data[data_key] for data_key in data_keys_without_fields}

        mapped_data = {
            mapped_key: mapped_value
            for mapped_key, mapped_value in mapped_data.items()
            if mapped_key in dataclass_fields
        }

    required_field_names = [field_name for field_name, field in dataclass_fields.items() if _field_is_required(field)]

    missing_required_fields = set(required_field_names) - set(mapped_data)
    if missing_required_fields:
        raise InvalidDeserializationException(f"Data is missing required fields: {missing_required_fields}")

    from any_serde import serde

    casted_data = {
        field_name: serde.from_data(field_types[field_name], field_data)
        for field_name, field_data in mapped_data.items()
    }

    obj = type_(**casted_data)

    if injected_data is not None:
        setattr(obj, ATTR_INJECTED_DATA, injected_data)

    return obj


def to_data(type_: Type[T_Dataclass], item: T_Dataclass) -> JSON:
    type_ = type(item)

    assert is_dataclass_type(type_), "Can only call dataclass_serde.to_data on dataclass instances!"

    field_types = _get_type_hints(type_)  # type: ignore
    dataclass_field_names = {x.name for x in dataclasses.fields(type_)}  # type: ignore

    fields_without_types = set(dataclass_field_names) - set(field_types)
    assert not fields_without_types, f"Fields without types: {fields_without_types}"

    serialization_renames = _get_serialization_renames(type_)  # type: ignore[arg-type]

    from any_serde import serde

    data = {
        serialization_renames.get(field_name, field_name): serde.to_data(
            field_types[field_name], getattr(item, field_name)
        )
        for field_name in dataclass_field_names
    }

    if hasattr(item, ATTR_INJECTED_DATA):
        injected_data = getattr(item, ATTR_INJECTED_DATA)
        assert isinstance(injected_data, dict)
        assert not set(injected_data) & set(data), "Trying to inject over existing keys"
        data.update(injected_data)

    return data


def register_serialization_renames(serialization_renames: Dict[str, str], type_: T_Dataclass) -> None:
    """Allows the serialization for a dataclass to have different keys than the dataclass."""
    # TODO: validate serialization renames here as well
    setattr(type_, ATTR_SERIALIZATION_RENAMES, serialization_renames)


@dataclasses.dataclass
class UnknownDataKeysPolicy:
    allow: bool
    roundtrip: bool


def allow_unknown_data_keys(type_: T_Dataclass, *, roundtrip: bool) -> None:
    policy = UnknownDataKeysPolicy(
        allow=True,
        roundtrip=roundtrip,
    )
    setattr(type_, ATTR_UNKNOWN_DATA_KEYS_POLICY, policy)


def dataclass_from_environ(
    dataclass_type: Type[T_Dataclass],
    environ: Optional[os._Environ] = None,
) -> T_Dataclass:
    if environ is None:
        environ = os.environ

    field_types = _get_type_hints(dataclass_type)  # type: ignore

    dataclass_args: Dict[str, JSON] = {}
    for field in dataclasses.fields(dataclass_type):  # type: ignore[arg-type]
        field_env_name = field.name.upper()
        if field_env_name not in environ:
            if _field_is_required(field):
                raise ValueError(f"Env var not set: {field_env_name}")
            continue

        raw_value: str = environ[field_env_name]
        value_type = field_types[field.name]

        if issubclass(value_type, str):
            parsed_value: JSON = raw_value
        elif issubclass(value_type, bool):
            parsed_value = bool(raw_value)
        elif issubclass(value_type, float):
            try:
                parsed_value = float(raw_value)
            except ValueError as err:
                raise InvalidDeserializationException from err
        elif issubclass(value_type, int):
            try:
                float_value = float(raw_value)
            except ValueError as err:
                raise InvalidDeserializationException from err
            int_value = int(float_value)
            if float_value != int_value:
                raise ValueError(f"Could not convert {raw_value} to int!")
            parsed_value = int_value
        else:
            raise ValueError(f"Cannot parse {value_type} from env vars yet!")

        dataclass_args[field.name] = parsed_value

    return dataclass_type(**dataclass_args)



from .serde import (
    from_data,
    to_data,
)
from .common import (
    JSON,
    InvalidDeserializationException,
    InvalidSerializationException,
)
from .dataclass_serde import (
    dataclass_from_environ,
    register_serialization_renames,
    allow_unknown_data_keys,
)

__all__ = [
    "from_data",
    "to_data",
    "JSON",
    "InvalidDeserializationException",
    "InvalidSerializationException",
    "dataclass_from_environ",
    "register_serialization_renames",
    "allow_unknown_data_keys",
]



from enum import Enum
from typing import Type, TypeVar, Any
from any_serde.common import InvalidDeserializationException, JSON

T_Enum = TypeVar("T_Enum", bound=Enum)


def is_enum_type(typ: Any) -> bool:
    return isinstance(typ, type) and issubclass(typ, Enum)


def from_data(type_: Type[T_Enum], data: JSON) -> T_Enum:
    if not isinstance(data, str):
        raise InvalidDeserializationException(f"Enums serialize to strings. Got {type(data)} instead!")

    split_items = data.split(".")

    if len(split_items) != 2:
        raise InvalidDeserializationException(f"Serialized enums are EnumType.ENUM_VALUE. Got {data} instead!")

    enum_type_str, enum_value_str = split_items

    enum_value = type_[enum_value_str]

    if not data == str(enum_value):
        raise InvalidDeserializationException(f"Wrong Enum type? Expected {type_}, got {enum_type_str}!")

    return enum_value


def to_data(enum_value: Enum) -> str:
    return str(enum_value)



from typing import Type, TypeVar, Any

from any_serde.common import (
    JSON_PRIMITIVE_TYPES,
    JSONPrimitive,
    InvalidDeserializationException,
    JSON,
)

T_JSONPrimitive = TypeVar("T_JSONPrimitive", bound=JSONPrimitive)


def is_primitive_type(typ: Any) -> bool:
    return typ in JSON_PRIMITIVE_TYPES


def from_data(type_: Type[T_JSONPrimitive], data: JSON) -> T_JSONPrimitive:
    assert is_primitive_type(type_), f"Can only call primitives_serde.from_data on primitive types! Got {type_}"

    if type_ is float and isinstance(data, int):
        return float(data)  # type: ignore

    if type_ is int and isinstance(data, float) and data % 1 == 0:
        return round(data)  # type: ignore

    if type_ is None and data is None:
        return data

    if type(data) is not type_:
        raise InvalidDeserializationException(f"Cannot deserialize {data} (type={type(data)}) to {type_}!")

    return data  # type: ignore


def to_data(item: JSONPrimitive) -> JSON:
    return item



from setuptools import find_packages, setup

setup(
    name="any_serde",
    version="0.2.3",
    packages=find_packages(),
    include_package_data=True,
    install_requires=[],
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    package_data={
        "": ["*py.typed"],
    },
)



from typing import Union, TypeVar, Type, Dict, List


JSON_PRIMITIVE_TYPES = [
    str,
    int,
    float,
    bool,
    None,
    type(None),
]
JSONPrimitive = Union[
    str,
    int,
    float,
    bool,
    None,
]
JSON = Union[
    JSONPrimitive,
    Dict[str, "JSON"],
    List["JSON"],
]


T_Any = TypeVar("T_Any")


def resolve_newtypes(typ: Type[T_Any]) -> Type[T_Any]:
    # https://github.com/python/mypy/issues/3325
    while hasattr(typ, "__supertype__"):
        typ = getattr(typ, "__supertype__")
    return typ


class InvalidDeserializationException(Exception):
    """Indicates that the given data is not valid for the target type given."""


class InvalidSerializationException(Exception):
    """Indicates that the given variable could not be serialized."""



import types
from typing import (
    Any,
    Sequence,
    Type,
    TypeVar,
    Union,
    get_args,
    get_origin,
)
from any_serde.common import InvalidSerializationException, InvalidDeserializationException, JSON, resolve_newtypes


T_Any = TypeVar("T_Any")


def _get_union_args(type_: Type[T_Any]) -> Sequence[Type[Any]]:
    type_origin_nullable = get_origin(type_)

    assert type_origin_nullable is not None, f"Calling union serde on non-union type: {type_}"

    type_origin = resolve_newtypes(type_origin_nullable)
    type_args = [resolve_newtypes(type_arg) for type_arg in get_args(type_)]

    assert type_origin in (Union, types.UnionType), f"Calling union serde on non-union type: {type_}"

    deduped_type_args = list(set(type_args))

    return deduped_type_args


def from_data(
    type_: Type[T_Any],
    data: JSON,
) -> T_Any:
    type_args = _get_union_args(type_)

    from any_serde import serde

    for union_arg in type_args:
        try:
            return serde.from_data(union_arg, data)
        except InvalidDeserializationException:
            pass

    raise InvalidDeserializationException("Could not deserialize to any union option.")


def to_data(
    type_: Type[T_Any],
    item: T_Any,
) -> JSON:
    type_args = _get_union_args(type_)

    from any_serde import serde

    for union_arg in type_args:
        try:
            return serde.to_data(union_arg, item)
        except InvalidSerializationException:
            pass

    raise InvalidSerializationException("Could not serialize from any union option.")



import types
from typing import (
    Any,
    Dict,
    Literal,
    Type,
    TypeVar,
    Union,
    get_args,
    get_origin,
)
from any_serde.common import (
    InvalidSerializationException,
    InvalidDeserializationException,
    JSON,
    resolve_newtypes,
)
from any_serde import enum_serde, primitives_serde, dataclass_serde, union_serde


T_Any = TypeVar("T_Any")


def from_data(
    type_: Type[T_Any],
    data: JSON,
) -> T_Any:
    """Constructs a python variable of the given type from JSON data.

    Args:
        _type: The type of the variable to construct. E.g. list[int] or (Set[str] | None)
        data: The JSON data to deserialize
    """
    type_ = resolve_newtypes(type_)

    if primitives_serde.is_primitive_type(type_):
        return primitives_serde.from_data(type_, data)  # type: ignore

    if enum_serde.is_enum_type(type_):
        return enum_serde.from_data(type_, data)  # type: ignore

    if dataclass_serde.is_dataclass_type(type_):
        return dataclass_serde.from_data(type_, data)

    type_origin_nullable = get_origin(type_)
    if type_origin_nullable is None:
        raise TypeError(f"Unsupported type: {type_}")

    type_origin = resolve_newtypes(type_origin_nullable)
    type_args = [resolve_newtypes(type_arg) for type_arg in get_args(type_)]

    if type_origin in (Union, types.UnionType):
        return union_serde.from_data(type_, data)

    if type_origin in (list, set):
        if not isinstance(data, list):
            raise InvalidDeserializationException(
                f"{type_origin} objects serialize to lists. Got {type(data)} instead!"
            )

        assert len(type_args) == 1
        list_item_type = type_args[0]

        return type_origin(
            (from_data(list_item_type, list_item) for list_item in data),
        )

    if type_origin is dict:
        if not isinstance(data, dict):
            raise InvalidDeserializationException(f"Dict objects serialize to dicts. Got {type(data)} instead!")

        assert len(type_args) == 2

        key_type, value_type = type_args

        return {from_data(key_type, key): from_data(value_type, value) for key, value in data.items()}  # type: ignore

    if type_origin is tuple:
        assert type_args

        if not isinstance(data, list):
            raise InvalidDeserializationException(f"Tuple objects serialize to lists. Got {type(data)} instead!")

        if len(data) != len(type_args):
            raise InvalidDeserializationException(
                f"Tuple type length {len(type_args)} does not match data length {len(data)}!"
            )

        return tuple(  # type: ignore
            from_data(tuple_item_type, tuple_item) for tuple_item_type, tuple_item in zip(type_args, data)
        )

    if type_origin is Literal:
        assert type_args

        if len(type_args) > 1:
            return from_data(
                Union[  # type: ignore
                    tuple(Literal[type_arg] for type_arg in type_args)  # type: ignore
                ],
                data,
            )

        literal_value = type_args[0]

        # TODO: figure out how to pass serialization options
        item = from_data(type(literal_value), data)

        assert type(item) == type(literal_value)

        if item != literal_value:
            raise InvalidSerializationException("Deserialized value does not match Literal type!")

        return item  # type: ignore

    raise NotImplementedError(f"Unsupported type_origin: {type_origin}")


def to_data(type_: Type[T_Any], item: T_Any) -> JSON:
    """Converts a python variable to JSON data.

    Args:
        type_: The intended type of item
        item: The python variable to serialize
    """
    type_ = resolve_newtypes(type_)

    if primitives_serde.is_primitive_type(type_):
        # TODO check type matches item
        return primitives_serde.to_data(item)  # type: ignore

    if enum_serde.is_enum_type(type_):
        return enum_serde.to_data(item)  # type: ignore

    if dataclass_serde.is_dataclass_type(type_):
        if type(item) is not type_:
            raise InvalidSerializationException(f"Specified type_ {type_} does not match item type {type(item)}!")
        return dataclass_serde.to_data(type_, item)

    type_origin_nullable = get_origin(type_)
    if type_origin_nullable is None:
        raise InvalidSerializationException(f"Unsupported type: {type_}")

    type_origin = resolve_newtypes(type_origin_nullable)
    type_args = [resolve_newtypes(type_arg) for type_arg in get_args(type_)]

    if type_origin in (Union, types.UnionType):
        return union_serde.to_data(type_, item)

    if type_origin in (list, set):
        if not isinstance(item, type_origin):
            raise InvalidSerializationException(f"Specified type_ {type_} does not match item type {type(item)}!")

        assert len(type_args) == 1
        list_item_type = type_args[0]

        return type_origin(
            (to_data(list_item_type, list_item) for list_item in item),
        )

    if type_origin is dict:
        if not isinstance(item, dict):
            raise InvalidSerializationException(f"Specified type_ {type_} does not match item type {type(item)}!")

        assert len(type_args) == 2

        key_type, value_type = type_args

        data: Dict[str, Any] = {}

        for key, value in item.items():
            key_data = to_data(key_type, key)
            if not isinstance(key_data, str):
                raise InvalidSerializationException(
                    f"Dict keys must serialize to strings. Got {type(key_data)} instead!"
                )
            value_data = to_data(value_type, value)
            data[key_data] = value_data

        return data

    if type_origin is tuple:
        assert type_args
        if not isinstance(item, tuple):
            raise InvalidSerializationException(f"Specified type_ {type_} does not match item type {type(item)}!")

        if len(item) != len(type_args):
            raise InvalidSerializationException(
                f"Specified type {type_} has length {len(type_args)} but item has {len(item)}!"
            )

        return [to_data(tuple_item_type, tuple_item) for tuple_item_type, tuple_item in zip(type_args, item)]

    if type_origin is Literal:
        assert type_args

        if len(type_args) > 1:
            return to_data(
                Union[  # type: ignore
                    tuple(Literal[type_arg] for type_arg in type_args)  # type: ignore
                ],
                item,
            )

        literal_value = type_args[0]

        if type(item) != type(literal_value):
            raise InvalidSerializationException(
                f"Item type {type(item)} does not match literal type {type(literal_value)}!"
            )

        if item != literal_value:
            raise InvalidSerializationException(f"Item {item} does not match literal value {literal_value}!")

        return to_data(type(literal_value), item)

    raise NotImplementedError(f"Unsupported type_origin: {type_origin}")
